/*
  $Id: Skyline.h 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#ifndef _Skyline_
#define _Skyline_

#include "Dataset.h"

//--

template <class Tuple> 
bool dominates(const Tuple &tupleX, const Tuple &tupleY)
// assumption: will be used only on tuples form the second relation
{
  return
    tupleX.relaxA() <= tupleY.relaxA() && 

#ifdef DIM_X_2
    tupleX.relaxB() <= tupleY.relaxB() &&
#endif

#ifdef DIM_X_3
    tupleX.relaxC() <= tupleY.relaxC() &&
#endif

    (tupleX.relaxA() <  tupleY.relaxA()

#ifdef DIM_X_2
     || tupleX.relaxB() <  tupleY.relaxB()
#endif

#ifdef DIM_X_3
     || tupleX.relaxC() <  tupleY.relaxC()
#endif

     );
}

template <class TupleR, class TupleS> 
bool dominates(const TupleR &tupleRX, const TupleS &tupleSX,
	       const TupleR &tupleRY, const TupleS &tupleSY)
{
  return
    tupleRX.relaxA() <= tupleRY.relaxA() && 

#ifdef DIM_2_X
    tupleRX.relaxB() <= tupleRY.relaxB() &&
#endif

#ifdef DIM_3_X
    tupleRX.relaxC() <= tupleRY.relaxC() &&
#endif

    tupleSX.relaxA() <= tupleSY.relaxA() && 

#ifdef DIM_X_2
    tupleSX.relaxB() <= tupleSY.relaxB() &&
#endif

#ifdef DIM_X_3
    tupleSX.relaxC() <= tupleSY.relaxC() &&
#endif

    (tupleRX.relaxA() <  tupleRY.relaxA()

#ifdef DIM_2_X     
     || tupleRX.relaxB() <  tupleRY.relaxB()
#endif

#ifdef DIM_3_X     
     || tupleRX.relaxC() <  tupleRY.relaxC()
#endif

     || tupleSX.relaxA() <  tupleSY.relaxA()

#ifdef DIM_X_2
     || tupleSX.relaxB() <  tupleSY.relaxB()
#endif

#ifdef DIM_X_3
     || tupleSX.relaxC() <  tupleSY.relaxC()
#endif

     );  
}

//--

template <class Tuple> 
bool dominates(const vector<Tuple> &dataset, const VectUnsigned &sky, const unsigned i) 
{
  for (VectUnsigned::const_iterator it = sky.begin(); it != sky.end(); ++it)
    {
      if (dominates(dataset[*it], dataset[i]))
	{
	  return true;
	}
    }
  return false;
}

template <class TupleR, class TupleS> 
bool dominates(const vector<TupleR> &datasetR, const  vector<TupleS> &datasetS,
	       const VectPairUnsigned &sky, const PairUnsigned p) 
{
  for (VectPairUnsigned::const_iterator it = sky.begin(); it != sky.end(); ++it)
    {
      if (dominates(datasetR[it->first], datasetS[it->second], datasetR[p.first], datasetS[p.second]))
	{
	  return true;
	}
    }
  return false;
}

template <class Tuple>
void removeDominated(const vector<Tuple> &dataset, VectUnsigned &sky, const unsigned i) 
{
  for (VectUnsigned::iterator it = sky.begin(); it != sky.end(); )
    {
      if (dominates(dataset[i], dataset[*it]))
	{
 	  sky.erase(it);
	}
      else
	{
	  ++it;
	}
    }
}

template <class TupleR, class TupleS>
void removeDominated(const vector<TupleR> &datasetR, const  vector<TupleS> &datasetS,
		     VectPairUnsigned &sky, const PairUnsigned p) 
{
  for (VectPairUnsigned::iterator it = sky.begin(); it != sky.end(); )
    {
      if (dominates(datasetR[p.first], datasetS[p.second], datasetR[it->first], datasetS[it->second]))
	{
 	  sky.erase(it);
	}
      else
	{
	  ++it;
	}
    }
}

//-- Insert Skyline

// Local Skyline
template <class Tuple>
void insertSkyline(const vector<Tuple> &dataset, 
		   VectUnsigned &sky, 
		   const unsigned id)
{
#ifdef TopK
  bool to_insert = false;
  if (sky.size() < TopK) 
    {
      to_insert = true;
    }
  else if (dataset[id].score() < dataset[sky.back()].score()) 
    {
      sky.pop_back();
      to_insert = true;
    }
  if (to_insert) 
    {
      VectUnsigned::iterator it = sky.begin();
      for ( ; it != sky.end() && dataset[id].score() >= dataset[*it].score(); ++it);
      sky.insert(it, id);
    }
#else
  sky.push_back(id);
#endif
}

// Global Skyline
template <class TupleR, class TupleS>
void insertSkyline(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
		   VectPairUnsigned &sky, 
		   const PairUnsigned &id)
{
#ifdef TopK
  bool to_insert = false;
  if (sky.size() < TopK) 
    {
      to_insert = true;
    }
  else if ((datasetR[id.first].score() + datasetS[id.second].score())
	   < (datasetR[sky.back().first].score() + datasetS[sky.back().second].score())) 
    {
      sky.pop_back();
      to_insert = true;
    }
  if (to_insert) 
    {
      VectPairUnsigned::iterator it = sky.begin();
      for ( ; it != sky.end() &&
	      (datasetR[id.first].score() + datasetS[id.second].score())
	      >= (datasetR[it->first].score() + datasetS[it->second].score()); ++it);
      sky.insert(it, id);
    }
#else
  sky.push_back(id);
#endif
}

//-- LOCAL

template <class Tuple> 
void localSkyline(const vector<Tuple> &dataset,
		  const PairHashIdxUnsignedCIt idxS,
		  VectUnsigned &sky)
{
  for (HashIdxUnsigned::const_iterator it = idxS.first; it != idxS.second; ++it) 
    {
      if (!dominates(dataset, sky, it->second)) 
	{
	  removeDominated(dataset, sky, it->second);
	  insertSkyline(dataset, sky, it->second);
	}
    }
}

//-- GLOBAL

template <class TupleR, class TupleS> 
void globalSkyline(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
		   const unsigned i, const VectUnsigned &localSky,
		   VectPairUnsigned &globalSky)
{
  for (VectUnsigned::const_iterator it = localSky.begin(); it != localSky.end(); ++it) 
    {
      PairUnsigned p = PairUnsigned(i, *it);
      if (!dominates(datasetR, datasetS, globalSky, p)) 
	{
	  removeDominated(datasetR, datasetS, globalSky, p);
	  insertSkyline(datasetR, datasetS, globalSky, p);
	}
    }
}

template <class TupleR, class TupleS> 
void globalSkyline(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
		   const unsigned i, const PairHashIdxUnsignedCIt idxS,
		   VectPairUnsigned &globalSky)
{
  for (HashIdxUnsignedCIt it = idxS.first; it != idxS.second; ++it) 
    {
      PairUnsigned p = PairUnsigned(i, it->second);
      if (!dominates(datasetR, datasetS, globalSky, p)) 
	{
	  removeDominated(datasetR, datasetS, globalSky, p);
	  insertSkyline(datasetR, datasetS, globalSky, p);
	}
    }
}

template <class TupleR, class TupleS, class Container> 
void globalSkyline(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
		   const Container &joinRS,
		   VectPairUnsigned &globalSky)
{
  for (typename Container::const_iterator it = joinRS.begin(); it != joinRS.end(); ++it) 
    {
      if (!dominates(datasetR, datasetS, globalSky, *it)) 
	{
	  removeDominated(datasetR, datasetS, globalSky, *it);
	  insertSkyline(datasetR, datasetS, globalSky, *it);
	}
    }
}

#endif
